#include "fancontroled.h"

static char *pid_filename = NULL;
static fancontroled_data fancontroled_d;

static void print_help(void) {
#ifndef STRIP_ALL
	printf("Usage: %s [OPTIONS] \n", FANCONTROLED_PROGRAM);
	printf(" -t  --temp-sensor-chip   Temperature sensor chip (e.g. lm75) \n");
	printf(" -r  --temp-read-field    Temperature read fild (e.g. temp1_input) \n");
	printf(" -m  --min-temp           Min fan speed on this temperature (°C) \n");
	printf(" -M  --max-temp           Max fan speed on this temperature (°C) \n");
	printf(" -f  --fan-sensor-chip    Fan sensor chip (e.g. g760a) \n");
	printf(" -c  --fan-control-field  Fan control field (e.g. pwm1) \n");
	printf(" -p  --pwm-fan-min        Min fan PWM \n");
	printf(" -P  --pwm-fan-max        Max fan PWM \n");
	printf(" -z  --pid-filename       PID filename \n");
	printf(" -v, --version            Display the program version \n");
	printf(" -h, --help               Display this help text \n");
#endif
}

static void print_version(void) {
#ifndef STRIP_ALL
	printf("%s version %s with libsensors version %s\n", FANCONTROLED_PROGRAM, FANCONTROLED_VERSION, libsensors_version);
#endif
}

static void signal_handler(int sig) {
	switch (sig) {
	case SIGINT:
	case SIGHUP:
	case SIGQUIT:
	case SIGTERM:
		if (set_fan_speed(&fancontroled_d, FANCONTROLED_PWM_FAN_MAX) != FANCONTROLED_RETURN_SUCCESS) {
			exit(EXIT_FAILURE);
		}
		sensors_cleanup();
		if (strlen(pid_filename)) {
			if (!unlink(pid_filename)) {
				exit(EXIT_FAILURE);
			}
		}
		exit(EXIT_SUCCESS);
		break;
	default:
		break;
	}
}

static uint8_t fancontroled_sensors_control(struct fancontroled_data *fancontroled_d_ptr) {
	uint8_t err;
	double temp;
	while (1) {
		temp = FANCONTROLED_NONEXISTENT_TEMPERATURE;
		if ((err = get_temp(fancontroled_d_ptr, &temp)) != FANCONTROLED_RETURN_SUCCESS) {
#ifdef DEV_DEBUG
			fprintf(stderr, "get_temp err value: %d \n", err);
#endif
			return err;
		}
#ifdef DEV_DEBUG
		printf("extracted temp value: %f \n", temp);
#endif

		uint8_t pwm;
		if (temp == FANCONTROLED_NONEXISTENT_TEMPERATURE) {
			pwm = FANCONTROLED_PWM_FAN_MAX;
		} else {
			if (((uint8_t) temp) < (fancontroled_d_ptr->temp_min)) {
				pwm = FANCONTROLED_DATA_PWM_FAN_OFF;
			}
			if (((uint8_t) temp) >= (fancontroled_d_ptr->temp_min) && ((uint8_t) temp) <= (fancontroled_d_ptr->temp_max)) {
				double tmp_temp;
				tmp_temp = temp - (double)(fancontroled_d_ptr->temp_min);
				tmp_temp = tmp_temp / (double)((fancontroled_d_ptr->temp_max) - (fancontroled_d_ptr->temp_min));
				uint8_t tmp_pwm;
				tmp_pwm = (fancontroled_d_ptr->pwm_fan_max) - (fancontroled_d_ptr->pwm_fan_min);
				pwm = (uint8_t) ((double)tmp_temp * (double)tmp_pwm) + (fancontroled_d_ptr->pwm_fan_min);
			}
			if (((uint8_t) temp) > (fancontroled_d_ptr->temp_max)) {
				pwm = FANCONTROLED_PWM_FAN_MAX;
			}
		}

#ifdef DEV_DEBUG
		printf("calculated pwm value: %d \n", pwm);
#endif

		if ((err = set_fan_speed(fancontroled_d_ptr, pwm)) != FANCONTROLED_RETURN_SUCCESS) {
#ifdef DEV_DEBUG
			fprintf(stderr, "set_fan_speed err value: %d \n", err);
#endif
			return err;
		}

		sleep(FANCONTROLED_SLEEP_TIME);
	}
	return FANCONTROLED_RETURN_SUCCESS;
}

int main(int argc, char *argv[]) {
	setlocale(LC_CTYPE, "");

	signal(SIGINT, signal_handler);
	signal(SIGHUP, signal_handler);
	signal(SIGQUIT, signal_handler);
	signal(SIGTERM, signal_handler);

	if (getuid() != 0) {
#ifndef STRIP_ALL
		fprintf(stderr, "run as root, exiting! \n");
#endif
		exit(EXIT_FAILURE);
	}

	if (sensors_init(NULL)) {
#ifndef STRIP_ALL
		fprintf(stderr, "could not initialize sensors! \n");
#endif
		return E_FANCONTROLED_SENSORS;
	}

	struct option long_opts[] = {
		{"temp-sensor-chip", required_argument, NULL, 't'},
		{"temp-read-field", required_argument, NULL, 'r'},
		{"temp-min", required_argument, NULL, 'm'},
		{"temp-max", required_argument, NULL, 'M'},
		{"fan-sensor-chip", required_argument, NULL, 'f'},
		{"fan-control-field", required_argument, NULL, 'c'},
		{"pwm-fan-min", required_argument, NULL, 'p'},
		{"pwm-fan-max", required_argument, NULL, 'P'},
		{"pid-filename", required_argument, NULL, 'z'},
		{"version", no_argument, NULL, 'v'},
		{"help", no_argument, NULL, 'h'},
		{NULL, 0, NULL, 0}
	};

	int c;
	bool temp_sensor_chip_arg_treated = false;
	bool fan_sensor_chip_arg_treated = false;
	while (1) {
		c = getopt_long(argc, argv, "t:r:m:M:f:c:p:P:z:hv", long_opts, NULL);
		if (c == EOF)
			break;
		switch (c) {
		case 't':
			{
				const sensors_chip_name *chip_name = NULL;
				int chip_nr = 0;
				uint8_t err;
				err = E_FANCONTROLED_FAILURE;
				while ((chip_name = sensors_get_detected_chips(NULL, &chip_nr)) != NULL) {
					if (memcmp(chip_name->prefix, optarg, strlen(optarg)) == 0) {
						fancontroled_d.temp_sensor = *chip_name;
						err = FANCONTROLED_RETURN_SUCCESS;
						break;
					}
				}
				if (err != FANCONTROLED_RETURN_SUCCESS) {
#ifndef STRIP_ALL
					fprintf(stderr, "could not find temp chip: %s! \n", optarg);
#endif
					exit(EXIT_FAILURE);
#ifdef DEV_DEBUG
				} else {
					printf("found temp sensor: %s \n", fancontroled_d.temp_sensor.prefix);
#endif
				}
				temp_sensor_chip_arg_treated = true;
			}
			break;
		case 'r':
			{
				if (!temp_sensor_chip_arg_treated) {
#ifndef STRIP_ALL
					fprintf(stderr, "define temp sensor chip before temp read field! \n");
#endif
					exit(EXIT_FAILURE);
				}
#ifdef DEV_DEBUG
				printf("temp read field: %s \n", optarg);
#endif
				const sensors_chip_name *chip_name = &fancontroled_d.temp_sensor;
				const sensors_feature *chip_feat_name = NULL;
				int chip_feat_nr = 0;
				const sensors_subfeature *chip_subfeat_name = NULL;
				int chip_subfeat_nr = 0;
				int err;
				char *label;
				double val;
				chip_feat_nr = 0;
				while ((chip_feat_name = sensors_get_features(chip_name, &chip_subfeat_nr))) {
					if (!(label = sensors_get_label(chip_name, chip_feat_name))) {
						continue;
					}
#ifdef DEV_DEBUG
					printf(" temp label: %s \n", label);
#endif
					free(label);
					chip_subfeat_nr = 0;
					while ((chip_subfeat_name = sensors_get_all_subfeatures(chip_name, chip_feat_name, &chip_subfeat_nr))) {
						if (chip_subfeat_name->flags & SENSORS_MODE_R) {
							if (memcmp(chip_subfeat_name->name, optarg, strlen(optarg)) == 0) {
#ifdef DEV_DEBUG
								printf(" temp subfeature number: %d \n", chip_subfeat_nr);
								printf(" temp subfeature name: %s \n", chip_subfeat_name->name);
#endif
								if (!sensors_get_value(chip_name, chip_subfeat_name->number, &val)) {
									fancontroled_d.temp_sensor_chip_subfeatature_number = chip_subfeat_nr - 1;
#ifdef DEV_DEBUG
									printf(" temp subfeature value: %.3f \n", val);
#endif
								}
							}
						}
					}
				}
			}
			break;
		case 'm':
			{
				int temp_min;
				temp_min = atoi(optarg);
				if ((temp_min >= 0) && (temp_min <= 255)) {
					fancontroled_d.temp_min = temp_min;
#ifdef DEV_DEBUG
					printf("temp min field: %d \n", fancontroled_d.temp_min);
#endif
				} else {
#ifndef STRIP_ALL
					fprintf(stderr, "temp min field out of range [0, 255]: %d! \n", temp_min);
#endif
					exit(EXIT_FAILURE);
				}
			}
			break;
		case 'M':
			{
				int temp_max;
				temp_max = atoi(optarg);
				if ((temp_max >= 0) && (temp_max <= 255)) {
					fancontroled_d.temp_max = temp_max;
#ifdef DEV_DEBUG
					printf("temp max field: %d \n", fancontroled_d.temp_max);
#endif
				} else {
#ifndef STRIP_ALL
					fprintf(stderr, "temp max field out of range [0, 255]: %d! \n", temp_max);
#endif
					exit(EXIT_FAILURE);
				}
			}
			break;
		case 'f':
			{
				const sensors_chip_name *chip_name = NULL;
				int chip_nr = 0;
				uint8_t err;
				err = E_FANCONTROLED_FAILURE;
				while ((chip_name = sensors_get_detected_chips(NULL, &chip_nr)) != NULL) {
					if (memcmp(chip_name->prefix, optarg, strlen(optarg)) == 0) {
						fancontroled_d.fan_sensor = *chip_name;
						err = FANCONTROLED_RETURN_SUCCESS;
						break;
					}
				}
				if (err != FANCONTROLED_RETURN_SUCCESS) {
#ifndef STRIP_ALL
					fprintf(stderr, "could not find fan chip %s! \n", optarg);
#endif
					exit(EXIT_FAILURE);
#ifdef DEV_DEBUG
				} else {
					printf("found fan sensor: %s \n", fancontroled_d.fan_sensor.prefix);
#endif
				}
				fan_sensor_chip_arg_treated = true;
			}
			break;
		case 'c':
			{
				if (!fan_sensor_chip_arg_treated) {
#ifndef STRIP_ALL
					fprintf(stderr, "define fan sensor chip before fan control field! \n");
#endif
					exit(EXIT_FAILURE);
				}
				if (strlen(optarg) >= FANCONTROLED_DATA_SENSORS_CHAR_ARRAY_MAX_LEN) {
#ifndef STRIP_ALL
					fprintf(stderr, "fan control field too long! \n");
#endif
					exit(EXIT_FAILURE);
				}
				if ((strlen(fancontroled_d.fan_sensor.path) + strlen(optarg)) >= NAME_MAX) {
#ifndef STRIP_ALL
					fprintf(stderr, "fan control field path too long! \n");
#endif
					exit(EXIT_FAILURE);
				}
#ifdef DEV_DEBUG
				printf("fan control field: %s \n", optarg);
#endif
				char fan_control_field_path[NAME_MAX];

#ifdef DEV_DEBUG
				printf(" fan sensor chip path: %s \n", fancontroled_d.fan_sensor.path);
#endif
				snprintf(fan_control_field_path, NAME_MAX, "%s/%s\0", fancontroled_d.fan_sensor.path, optarg);
#ifdef DEV_DEBUG
				printf(" calculated fan control field path: %s \n", fan_control_field_path);
#endif
				FILE *fan_control;
				if (fan_control = fopen(fan_control_field_path, "r")) {
					memset(fancontroled_d.fan_write_field, 0, FANCONTROLED_DATA_SENSORS_CHAR_ARRAY_MAX_LEN + 1);
					memcpy(fancontroled_d.fan_write_field, optarg, FANCONTROLED_DATA_SENSORS_CHAR_ARRAY_MAX_LEN);
#ifdef DEV_DEBUG
					printf(" fan control field exists: %s \n", fancontroled_d.fan_write_field);
					char line[FANCONTROLED_PWM_FAN_CONTROL_FILE_DATA_LENGTH + 1];
					memset(line, 0, FANCONTROLED_PWM_FAN_CONTROL_FILE_DATA_LENGTH + 1);
					fgets(line, FANCONTROLED_PWM_FAN_CONTROL_FILE_DATA_LENGTH, fan_control);
					printf(" fan control field value: %s \n", line);
#endif
					fclose(fan_control);
				} else {
#ifndef STRIP_ALL
					fprintf(stderr, "fan control field path does not exist! \n");
#endif
					exit(EXIT_FAILURE);
				}
			}
			break;
		case 'p':
			{
				int pwm_fan_min;
				pwm_fan_min = atoi(optarg);
				if ((pwm_fan_min >= 0) && (pwm_fan_min <= 255)) {
					fancontroled_d.pwm_fan_min = pwm_fan_min;
#ifdef DEV_DEBUG
					printf("min fan pwm field: %d \n", fancontroled_d.pwm_fan_min);
#endif
				} else {
#ifndef STRIP_ALL
					fprintf(stderr, "min fan pwm field out of range [0, 255]: %d! \n", pwm_fan_min);
#endif
					exit(EXIT_FAILURE);
				}
			}
			break;
		case 'P':
			{
				int pwm_fan_max;
				pwm_fan_max = atoi(optarg);
				if ((pwm_fan_max >= 0) && (pwm_fan_max <= 255)) {
					fancontroled_d.pwm_fan_max = pwm_fan_max;
#ifdef DEV_DEBUG
					printf("max fan pwm field: %d \n", fancontroled_d.pwm_fan_max);
#endif
				} else {
					fprintf(stderr, "max fan pwm field out of range [0, 255]: %d! \n", pwm_fan_max);
					exit(EXIT_FAILURE);
				}
			}
			break;
		case 'z':
			{
				if (strlen(optarg) < NAME_MAX) {
					pid_filename = (char *)malloc(sizeof(char) * strlen(optarg) + 1);
					if (pid_filename == NULL) {
#ifndef STRIP_ALL
						fprintf(stderr, "could not allocate memmory for filename! \n");
#endif
						exit(EXIT_FAILURE);
					}
					memset(pid_filename, 0, strlen(optarg) + 1);
					memcpy(pid_filename, optarg, strlen(optarg));
#ifdef DEV_DEBUG
					printf("pid file name field: %s \n", pid_filename);
#endif
				} else {
#ifndef STRIP_ALL
					fprintf(stderr, "pid filename too big: %s! \n", optarg);
#endif
					exit(EXIT_FAILURE);
				}
			}
			break;
		case 'h':
			print_help();
			exit(EXIT_SUCCESS);
		case 'v':
			print_version();
			exit(EXIT_SUCCESS);
		default:
#ifndef STRIP_ALL
			fprintf(stderr, "internal error while parsing options!\n");
#endif
			exit(EXIT_FAILURE);
		}
	}

	if (strlen(pid_filename) != 0) {
		FILE *f_read;
		if (f_read = fopen(pid_filename, "r")) {
#ifndef STRIP_ALL
			fprintf(stderr, "pid file exists: %s, exiting ...\n", pid_filename);
#endif
			exit(EXIT_FAILURE);
		}
	}

	pid_t pid;
	pid = fork();
	if (pid < 0) {
		exit(EXIT_FAILURE);
	}
	if (pid > 0) {
		exit(EXIT_SUCCESS);
	}

	pid_t sid;
	umask(027);

	sid = setsid();
	if (sid < 0) {
#ifndef STRIP_ALL
		fprintf(stderr, "setsid() function returned error! \n");
#endif
		exit(EXIT_FAILURE);
	}

	if ((chdir(FANCONTROLED_RUNNING_DIR)) < 0) {
		fprintf(stderr, "chdir() function returned error! \n");
		exit(EXIT_FAILURE);
	}

	FILE *f_write;
	if (f_write = fopen(pid_filename, "w")) {
		fprintf(f_write, "%d", getpid());
		fclose(f_write);
	}

#ifndef DEV_DEBUG
	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);
#endif

#ifdef DEV_DEBUG
	printf("== == == \n");
	printf("starting daemon loop... \n");
	printf("== == == \n");
#endif

	uint8_t val;
	if ((val = fancontroled_sensors_control(&fancontroled_d)) != FANCONTROLED_RETURN_SUCCESS) {
		exit(EXIT_FAILURE);
	}

	sensors_cleanup();
	exit(EXIT_SUCCESS);
}
